import test, { describe } from "node:test";
import assert from "node:assert";
import MarkdownDocument from "../document.js";
import Vault from "../vault.js";

function getTestDocument(): MarkdownDocument {
  const vault = new Vault("./bin/_test/test-vault");

  if (!vault.slugs.includes("testdoc")) {
    assert.fail("Could not find test document");
  }

  return vault.index["testdoc"]!;
}

describe("MarkdownDocument", () => {
  test("markdown and frontmatter are parsed out", () => {
    const doc = getTestDocument();
    assert.deepStrictEqual(doc.frontmatter, {
      property: "value",
      cssclasses: [],
      tags: [],
      aliases: [],
    });

    assert.equal(
      doc.markdown.trim(),
      "This is a test doc for use with testing docs.",
    );
  });

  test("changing the frontmatter updates the full document contents", () => {
    const doc = getTestDocument();
    const originalContents = doc.content;
    doc.setFrontmatter({
      ...doc.frontmatter,
      example: "hello",
    });

    assert.notEqual(doc.content, originalContents);
  });

  test("changing the markdown updates the full document contents", () => {
    const doc = getTestDocument();
    const originalContents = doc.content;
    doc.setMarkdown("Hello, worlt.");

    assert.notEqual(doc.content, originalContents);
  });

  test("setting the full content updates markdown and frontmatter", () => {
    const doc = getTestDocument();
    doc.setContent(`---
hello: world
---

Testing!
`);

    assert.deepStrictEqual(doc.frontmatter, {
      hello: "world",
      tags: [],
      aliases: [],
      cssclasses: [],
    });
    assert.equal(doc.markdown.trim(), "Testing!");
  });

  test("reverting works", () => {
    const doc = getTestDocument();
    const originalContents = doc.content;
    doc.setFrontmatter({
      ...doc.frontmatter,
      example: "hello",
    });
    doc.revert();

    assert.equal(doc.content, originalContents);
  });

  test("checking for tags", () => {
    const doc = getTestDocument();
    const taggedDoc = doc.vault.index["directory-subdirectory-i-have-tags"]!;
    assert.equal(taggedDoc.hasTag("tags"), true);
    assert.equal(doc.hasTag("tags"), false);
  });
  test("checking for taxonomy", () => {
    const doc = getTestDocument();
    const taggedDoc = doc.vault.index["directory-subdirectory-i-have-tags"]!;
    assert.equal(taggedDoc.hasTaxonomy(["Directory", "Subdirectory"]), true);
    assert.equal(doc.hasTaxonomy(["Directory", "Subdirectory"]), false);
  });
});
